'
' UTF-8.bi
'
' UTF-8 library for FreeBASIC, helper and conversion functions
'
' UTF8(...) - turns a list of (up to 8) integer values (Unicode) into a string of UTF8-Expressions
'                    returns empty string on error (invalid Unicode)
'
' UTF8FILL(INTEGER multiplier, ULONG Unicode) AS STRING - returns a string filled with
'                    the given Unicode char UTF-8 encoded repeated 'multiplier' times
'                    almost like STRING
'
' UNICODE(STRING substring[, INTEGER start]) - returns the Unicode value of the first UTF-8
'                    encoded character found in "substring", starts searching from "start"
'                    - first char is 1 (and/or 0)
'
' UTF8FROMISO8859(STRING string, INTEGER codetable=15) - converts an "ASCII" string into a
'                    UTF-8 string, using translation tables for the ISO/IEC 8859 standard
'                    from ISO 8859-1 to ISO 8859-16
'                    
' IS_VALIDUTF8(STRING string[, INTEGER start[, INTEGER end]]) - tests a given string whether
'                    all characters are valid UTF8 - returns TRUE if everything is all-right.
'                    This includes pure ASCII.
'                    Returns FALSE the first time any illegal code appears; prevents memory
'                    leaks
'
' UTF8LEN(STRING ) - returns length of a UTF-8 string in chars, rather than bytes
'
' UPPERCASE(STRING destination, INTEGER length=0) - tries to capitalise a UTF-8 string, turning
'                    all minuscules into majuscules, length=1 only capitalises the first letter
'
' LOWERCASE(STRING destination) - tries to turn a UTF-8 string into complete lower case,
'                    replacing all majuscules with their minuscules
'

DIM SHARED iso8859tableentry(16, 96) AS ULONG = _
           {{&H0A0, &H0A1, &H0A2, &H0A3, &H0A4, &H0A5, &H0A6, &H0A7, &H0A8, &H0A9, &H0AA, &H0AB, &H0AC, &H0AD, &H0AE, &H0AF, _
             &H0B0, &H0B1, &H0B2, &H0B3, &H0B4, &H0B5, &H0B6, &H0B7, &H0B8, &H0B9, &H0BA, &H0BB, &H0BC, &H0BD, &H0BE, &H0BF, _
             &H0C0, &H0C1, &H0C2, &H0C3, &H0C4, &H0C5, &H0C6, &H0C7, &H0C8, &H0C9, &H0CA, &H0CB, &H0CC, &H0CD, &H0CE, &H0CF, _
             &H0D0, &H0D1, &H0D2, &H0D3, &H0D4, &H0D5, &H0D6, &H0D7, &H0D8, &H0D9, &H0DA, &H0DB, &H0DC, &H0DD, &H0DE, &H0DF, _
             &H0E0, &H0E1, &H0E2, &H0E3, &H0E4, &H0E5, &H0E6, &H0E7, &H0E8, &H0E9, &H0EA, &H0EB, &H0EC, &H0ED, &H0EE, &H0EF, _
             &H0F0, &H0F1, &H0F2, &H0F3, &H0F4, &H0F5, &H0F6, &H0F7, &H0F8, &H0F9, &H0FA, &H0FB, &H0FC, &H0FD, &H0FE, &H0FF},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000},_
            {&H0A0, &H0A1, &H0A2, &H0A3,&H20AC, &H0A5, &H0A6, &H0A7, &H0A8, &H0A9, &H0AA, &H0AB, &H0AC, &H0AD, &H0AE, &H0AF, _
             &H0B0, &H0B1, &H0B2, &H0B3, &H0B4, &H0B5, &H0B6, &H0B7, &H0B8, &H0B9, &H0BA, &H0BB, &H0BC, &H0BD, &H0BE, &H0BF, _
             &H0C0, &H0C1, &H0C2, &H0C3, &H0C4, &H0C5, &H0C6, &H0C7, &H0C8, &H0C9, &H0CA, &H0CB, &H0CC, &H0CD, &H0CE, &H0CF, _
             &H0D0, &H0D1, &H0D2, &H0D3, &H0D4, &H0D5, &H0D6, &H0D7, &H0D8, &H0D9, &H0DA, &H0DB, &H0DC, &H0DD, &H0DE, &H0DF, _
             &H0E0, &H0E1, &H0E2, &H0E3, &H0E4, &H0E5, &H0E6, &H0E7, &H0E8, &H0E9, &H0EA, &H0EB, &H0EC, &H0ED, &H0EE, &H0EF, _
             &H0F0, &H0F1, &H0F2, &H0F3, &H0F4, &H0F5, &H0F6, &H0F7, &H0F8, &H0F9, &H0FA, &H0FB, &H0FC, &H0FD, &H0FE, &H0FF},_
            {&H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, _
             &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000, &H000} _
           }

FUNCTION UTF8(BYVAL code AS ULONG, BYVAL code1 AS ULONG=0, BYVAL code2 AS ULONG=0, BYVAL code3 AS ULONG=0, BYVAL code4 AS ULONG=0, BYVAL code5 AS ULONG=0, BYVAL code6 AS ULONG=0, BYVAL code7 AS ULONG=0) AS STRING
  DIM AS ULONG codeno(8), i
  DIM AS STRING result=""

  codeno(1)=code1 ' crude code for a list of Unicodes, allows to hand up to 8 values at once
  IF code1 THEN
    codeno(2)=code2
    codeno(3)=code3
    codeno(4)=code4
    codeno(5)=code5
    codeno(6)=code6
    codeno(7)=code7

    i=0
    WHILE code
      IF code<128 THEN
        ' ASCII range
        result+=CHR(code)
      ELSEIF code<&H0800 THEN
        ' 2 byte code
        result+=CHR(((code SHR 6) OR 192), (code AND 63)+128)
      ELSEIF code<&H10000 THEN
        ' 3 byte code
        result+=CHR(((code SHR 12) OR 224), ((code SHR 6) AND 63)+128, (code AND 63)+128)
      ELSEIF code<&H140000 THEN
        ' 4 byte code
        result+=CHR(((code SHR 18) OR 240), ((code SHR 12) AND 63)+128, ((code SHR 6) AND 63)+128, (code AND 63)+128)
      ELSE
        ' illegal code
        ' result=""
      ENDIF
      i+=1
      code=codeno(i)
    WEND
  ELSE ' slightly faster treatment if single value
    IF code<128 THEN
      ' ASCII range
      result+=CHR(code)
    ELSEIF code<&H0800 THEN
      ' 2 byte code
      result+=CHR(((code SHR 6) OR 192), (code AND 63)+128)
    ELSEIF code<&H10000 THEN
      ' 3 byte code
      result+=CHR(((code SHR 12) OR 224), ((code SHR 6) AND 63)+128, (code AND 63)+128)
    ELSEIF code<&H140000 THEN
      ' 4 byte code
      result+=CHR(((code SHR 18) OR 240), ((code SHR 12) AND 63)+128, ((code SHR 6) AND 63)+128, (code AND 63)+128)
    ELSE
      ' illegal code
      ' result=""
    ENDIF
  ENDIF

  RETURN result
END FUNCTION

FUNCTION UTF8FILL(BYVAL multiplier AS INTEGER, BYVAL unicode AS ULONG) AS STRING
  DIM AS STRING result="", ucode

  ucode=UTF8(unicode)

  WHILE multiplier
    result+=ucode
    multiplier-=1
  WEND

  RETURN result
END FUNCTION

FUNCTION UNICODE(BYVAL substr AS STRING, BYVAL start AS INTEGER=0) AS ULONG

END FUNCTION

FUNCTION UTF8FROMISO8859(BYVAL source AS string, BYVAL codetable AS integer=15) AS STRING
  ' iso8859tableentry()
END FUNCTION

FUNCTION IS_VALIDUTF8(BYVAL source AS STRING, BYVAL strstart AS INTEGER=0, BYVAL strend AS INTEGER=0) AS boolean

END FUNCTION

FUNCTION UTF8LEN(BYVAL source AS STRING) AS INTEGER

END FUNCTION

FUNCTION UPPERCASE(BYVAL source AS STRING, BYVAL length AS INTEGER=0) AS STRING

END FUNCTION

FUNCTION LOWERCASE(BYVAL source AS STRING) AS STRING

END FUNCTION
